@@ -1,6 +1,6 @@
 /*###########################################################################
  #
- # Copyright (c) 2003 Zope Corporation and Contributors.
+ # Copyright (c) 2003 Zope Foundation and Contributors.
  # All Rights Reserved.
  #
  # This software is subject to the provisions of the Zope Public License,
@@ -18,6 +18,16 @@
 #define TYPE(O) ((PyTypeObject*)(O))
 #define OBJECT(O) ((PyObject*)(O))
 #define CLASSIC(O) ((PyClassObject*)(O))
+#ifndef PyVarObject_HEAD_INIT
+#define PyVarObject_HEAD_INIT(a, b) PyObject_HEAD_INIT(a) b,
+#endif
+#ifndef Py_TYPE
+#define Py_TYPE(o) ((o)->ob_type)
+#endif
+
+#if PY_MAJOR_VERSION >= 3
+#define PY3K
+#endif
 
 static PyObject *str__dict__, *str__implemented__, *strextends;
 static PyObject *BuiltinImplementationSpecifications, *str__provides__;
@@ -32,7 +42,7 @@ static PyTypeObject *Implements;
 
 static int imported_declarations = 0;
 
-static int 
+static int
 import_declarations(void)
 {
   PyObject *declarations, *i;
@@ -40,7 +50,7 @@ import_declarations(void)
   declarations = PyImport_ImportModule("zope.interface.declarations");
   if (declarations == NULL)
     return -1;
-  
+
   BuiltinImplementationSpecifications = PyObject_GetAttrString(
                     declarations, "BuiltinImplementationSpecifications");
   if (BuiltinImplementationSpecifications == NULL)
@@ -62,7 +72,7 @@ import_declarations(void)
 
   if (! PyType_Check(i))
     {
-      PyErr_SetString(PyExc_TypeError, 
+      PyErr_SetString(PyExc_TypeError,
                       "zope.interface.declarations.Implements is not a type");
       return -1;
     }
@@ -131,7 +141,7 @@ implementedBy(PyObject *ignored, PyObject *cls)
   /* Maybe we have a builtin */
   if (imported_declarations == 0 && import_declarations() < 0)
     return NULL;
-  
+
   spec = PyDict_GetItem(BuiltinImplementationSpecifications, cls);
   if (spec != NULL)
     {
@@ -149,7 +159,7 @@ getObjectSpecification(PyObject *ignored, PyObject *ob)
   PyObject *cls, *result;
 
   result = PyObject_GetAttr(ob, str__provides__);
-  if (result != NULL)
+  if (result != NULL && PyObject_TypeCheck(result, &SpecType))
     return result;
 
   PyErr_Clear();
@@ -175,13 +185,13 @@ static PyObject *
 providedBy(PyObject *ignored, PyObject *ob)
 {
   PyObject *result, *cls, *cp;
-  
+
   result = PyObject_GetAttr(ob, str__providedBy__);
   if (result == NULL)
     {
       PyErr_Clear();
       return getObjectSpecification(NULL, ob);
-    } 
+    }
 
 
   /* We want to make sure we have a spec. We can't do a type check
@@ -189,11 +199,11 @@ providedBy(PyObject *ignored, PyObject *ob)
      only attribute.
   */
   if (PyObject_TypeCheck(result, &SpecType)
-      || 
+      ||
       PyObject_HasAttr(result, strextends)
       )
     return result;
-    
+
   /*
     The object's class doesn't understand descriptors.
     Sigh. We need to get an object descriptor, but we have to be
@@ -208,13 +218,13 @@ providedBy(PyObject *ignored, PyObject *ob)
 
   result = PyObject_GetAttr(ob, str__provides__);
   if (result == NULL)
-    {      
+    {
       /* No __provides__, so just fall back to implementedBy */
       PyErr_Clear();
       result = implementedBy(NULL, cls);
       Py_DECREF(cls);
       return result;
-    } 
+    }
 
   cp = PyObject_GetAttr(cls, str__provides__);
   if (cp == NULL)
@@ -241,9 +251,9 @@ providedBy(PyObject *ignored, PyObject *ob)
   return result;
 }
 
-/* 
+/*
    Get an attribute from an inst dict. Return a borrowed reference.
-  
+
    This has a number of advantages:
 
    - It avoids layers of Python api
@@ -269,7 +279,7 @@ inst_attr(PyObject *self, PyObject *name)
 
 static PyObject *
 Spec_extends(PyObject *self, PyObject *other)
-{  
+{
   PyObject *implied;
 
   implied = inst_attr(self, str_implied);
@@ -289,11 +299,11 @@ Spec_extends(PyObject *self, PyObject *other)
 #endif
 }
 
-static char Spec_extends__doc__[] = 
+static char Spec_extends__doc__[] =
 "Test whether a specification is or extends another"
 ;
 
-static char Spec_providedBy__doc__[] = 
+static char Spec_providedBy__doc__[] =
 "Test whether an interface is implemented by the specification"
 ;
 
@@ -320,7 +330,7 @@ Spec_providedBy(PyObject *self, PyObject *ob)
     item = Spec_extends(decl, self);
   else
     /* decl is probably a security proxy.  We have to go the long way
-       around. 
+       around.
     */
     item = PyObject_CallFunctionObjArgs(decl, self, NULL);
 
@@ -329,7 +339,7 @@ Spec_providedBy(PyObject *self, PyObject *ob)
 }
 
 
-static char Spec_implementedBy__doc__[] = 
+static char Spec_implementedBy__doc__[] =
 "Test whether the specification is implemented by a class or factory.\n"
 "Raise TypeError if argument is neither a class nor a callable."
 ;
@@ -342,7 +352,7 @@ Spec_implementedBy(PyObject *self, PyObject *cls)
   decl = implementedBy(NULL, cls);
   if (decl == NULL)
     return NULL;
-  
+
   if (PyObject_TypeCheck(decl, &SpecType))
     item = Spec_extends(decl, self);
   else
@@ -353,10 +363,10 @@ Spec_implementedBy(PyObject *self, PyObject *cls)
 }
 
 static struct PyMethodDef Spec_methods[] = {
-	{"providedBy",  
+	{"providedBy",
          (PyCFunction)Spec_providedBy,		METH_O,
 	 Spec_providedBy__doc__},
-	{"implementedBy", 
+	{"implementedBy",
          (PyCFunction)Spec_implementedBy,	METH_O,
 	 Spec_implementedBy__doc__},
 	{"isOrExtends",	(PyCFunction)Spec_extends,	METH_O,
@@ -366,8 +376,7 @@ static struct PyMethodDef Spec_methods[] = {
 };
 
 static PyTypeObject SpecType = {
-	PyObject_HEAD_INIT(NULL)
-	/* ob_size           */ 0,
+	PyVarObject_HEAD_INIT(NULL, 0)
 	/* tp_name           */ "_interface_coptimizations."
                                 "SpecificationBase",
 	/* tp_basicsize      */ 0,
@@ -376,7 +385,7 @@ static PyTypeObject SpecType = {
 	/* tp_print          */ (printfunc)0,
 	/* tp_getattr        */ (getattrfunc)0,
 	/* tp_setattr        */ (setattrfunc)0,
-	/* tp_compare        */ (cmpfunc)0,
+	/* tp_compare        */ 0,
 	/* tp_repr           */ (reprfunc)0,
 	/* tp_as_number      */ 0,
 	/* tp_as_sequence    */ 0,
@@ -414,8 +423,7 @@ OSD_descr_get(PyObject *self, PyObject *inst, PyObject *cls)
 }
 
 static PyTypeObject OSDType = {
-	PyObject_HEAD_INIT(NULL)
-	/* ob_size           */ 0,
+	PyVarObject_HEAD_INIT(NULL, 0)
 	/* tp_name           */ "_interface_coptimizations."
                                 "ObjectSpecificationDescriptor",
 	/* tp_basicsize      */ 0,
@@ -424,7 +432,7 @@ static PyTypeObject OSDType = {
 	/* tp_print          */ (printfunc)0,
 	/* tp_getattr        */ (getattrfunc)0,
 	/* tp_setattr        */ (setattrfunc)0,
-	/* tp_compare        */ (cmpfunc)0,
+	/* tp_compare        */ 0,
 	/* tp_repr           */ (reprfunc)0,
 	/* tp_as_number      */ 0,
 	/* tp_as_sequence    */ 0,
@@ -473,14 +481,13 @@ CPB_descr_get(PyObject *self, PyObject *inst, PyObject *cls)
       Py_XINCREF(implements);
       return implements;
     }
-  
+
   PyErr_SetObject(PyExc_AttributeError, str__provides__);
   return NULL;
 }
 
 static PyTypeObject CPBType = {
-	PyObject_HEAD_INIT(NULL)
-	/* ob_size           */ 0,
+	PyVarObject_HEAD_INIT(NULL, 0)
 	/* tp_name           */ "_interface_coptimizations."
                                 "ClassProvidesBase",
 	/* tp_basicsize      */ 0,
@@ -489,7 +496,7 @@ static PyTypeObject CPBType = {
 	/* tp_print          */ (printfunc)0,
 	/* tp_getattr        */ (getattrfunc)0,
 	/* tp_setattr        */ (setattrfunc)0,
-	/* tp_compare        */ (cmpfunc)0,
+	/* tp_compare        */ 0,
 	/* tp_repr           */ (reprfunc)0,
 	/* tp_as_number      */ 0,
 	/* tp_as_sequence    */ 0,
@@ -531,7 +538,7 @@ static PyTypeObject CPBType = {
             if adapter is not None:
                 return adapter
 
-  
+
 */
 static PyObject *
 __adapt__(PyObject *self, PyObject *obj)
@@ -560,7 +567,7 @@ __adapt__(PyObject *self, PyObject *obj)
   else
     {
       /* decl is probably a security proxy.  We have to go the long way
-         around. 
+         around.
       */
       PyObject *r;
       r = PyObject_CallFunctionObjArgs(decl, self, NULL);
@@ -608,7 +615,7 @@ static struct PyMethodDef ib_methods[] = {
   {NULL,		NULL}		/* sentinel */
 };
 
-/* 
+/*
         def __call__(self, obj, alternate=_marker):
             conform = getattr(obj, '__conform__', None)
             if conform is not None:
@@ -629,7 +636,7 @@ static PyObject *
 ib_call(PyObject *self, PyObject *args, PyObject *kwargs)
 {
   PyObject *conform, *obj, *alternate=NULL, *adapter;
-  
+
   static char *kwlist[] = {"obj", "alternate", NULL};
 
   if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O", kwlist,
@@ -648,7 +655,7 @@ ib_call(PyObject *self, PyObject *args, PyObject *kwargs)
     }
   else
     PyErr_Clear();
- 
+
   adapter = __adapt__(self, obj);
   if (adapter == NULL || adapter != Py_None)
     return adapter;
@@ -670,8 +677,7 @@ ib_call(PyObject *self, PyObject *args, PyObject *kwargs)
 }
 
 static PyTypeObject InterfaceBase = {
-	PyObject_HEAD_INIT(NULL)
-	/* ob_size           */ 0,
+	PyVarObject_HEAD_INIT(NULL, 0)
 	/* tp_name           */ "_zope_interface_coptimizations."
                                 "InterfaceBase",
 	/* tp_basicsize      */ 0,
@@ -680,7 +686,7 @@ static PyTypeObject InterfaceBase = {
 	/* tp_print          */ (printfunc)0,
 	/* tp_getattr        */ (getattrfunc)0,
 	/* tp_setattr        */ (setattrfunc)0,
-	/* tp_compare        */ (cmpfunc)0,
+	/* tp_compare        */ 0,
 	/* tp_repr           */ (reprfunc)0,
 	/* tp_as_number      */ 0,
 	/* tp_as_sequence    */ 0,
@@ -747,7 +753,7 @@ lookup_traverse(lookup *self, visitproc visit, void *arg)
     if (vret != 0)
       return vret;
   }
-  
+
   return 0;
 }
 
@@ -764,7 +770,7 @@ static void
 lookup_dealloc(lookup *self)
 {
   lookup_clear(self);
-  self->ob_type->tp_free((PyObject*)self);
+  Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
 /*
@@ -785,7 +791,7 @@ lookup_changed(lookup *self, PyObject *ignored)
                                         if (N == NULL) return NULL; \
                                        }
 
-/*  
+/*
     def _getcache(self, provided, name):
         cache = self._cache.get(provided)
         if cache is None:
@@ -808,7 +814,7 @@ _subcache(PyObject *cache, PyObject *key)
   if (subcache == NULL)
     {
       int status;
- 
+
       subcache = PyDict_New();
       if (subcache == NULL)
         return NULL;
@@ -837,7 +843,7 @@ _getcache(lookup *self, PyObject *provided, PyObject *name)
 }
 
 
-/*  
+/*
     def lookup(self, required, provided, name=u'', default=None):
         cache = self._getcache(provided, name)
         if len(required) == 1:
@@ -868,12 +874,12 @@ tuplefy(PyObject *v)
     }
   else
     Py_INCREF(v);
-  
+
   return v;
 }
 static PyObject *
-_lookup(lookup *self, 
-        PyObject *required, PyObject *provided, PyObject *name, 
+_lookup(lookup *self,
+        PyObject *required, PyObject *provided, PyObject *name,
         PyObject *default_)
 {
   PyObject *result, *key, *cache;
@@ -934,13 +940,13 @@ lookup_lookup(lookup *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &required, &provided, &name, &default_))
-    return NULL; 
+    return NULL;
 
   return _lookup(self, required, provided, name, default_);
 }
 
 
-/*  
+/*
     def lookup1(self, required, provided, name=u'', default=None):
         cache = self._getcache(provided, name)
         result = cache.get(required, _not_in_mapping)
@@ -953,8 +959,8 @@ lookup_lookup(lookup *self, PyObject *args, PyObject *kwds)
         return result
 */
 static PyObject *
-_lookup1(lookup *self, 
-        PyObject *required, PyObject *provided, PyObject *name, 
+_lookup1(lookup *self,
+        PyObject *required, PyObject *provided, PyObject *name,
         PyObject *default_)
 {
   PyObject *result, *cache;
@@ -977,7 +983,13 @@ _lookup1(lookup *self,
       Py_DECREF(tup);
     }
   else
-    Py_INCREF(result);
+    {
+      if (result == Py_None && default_ != NULL)
+        {
+          result = default_;
+        }
+      Py_INCREF(result);
+    }
 
   return result;
 }
@@ -989,12 +1001,12 @@ lookup_lookup1(lookup *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &required, &provided, &name, &default_))
-    return NULL; 
+    return NULL;
 
   return _lookup1(self, required, provided, name, default_);
 }
 
-/*  
+/*
     def adapter_hook(self, provided, object, name=u'', default=None):
         required = providedBy(object)
         cache = self._getcache(provided, name)
@@ -1010,8 +1022,8 @@ lookup_lookup1(lookup *self, PyObject *args, PyObject *kwds)
         return default
 */
 static PyObject *
-_adapter_hook(lookup *self, 
-              PyObject *provided, PyObject *object,  PyObject *name, 
+_adapter_hook(lookup *self,
+              PyObject *provided, PyObject *object,  PyObject *name,
               PyObject *default_)
 {
   PyObject *required, *factory, *result;
@@ -1019,12 +1031,12 @@ _adapter_hook(lookup *self,
   required = providedBy(NULL, object);
   if (required == NULL)
     return NULL;
-  
+
   factory = _lookup1(self, required, provided, name, Py_None);
   Py_DECREF(required);
   if (factory == NULL)
     return NULL;
-  
+
   if (factory != Py_None)
     {
       result = PyObject_CallFunctionObjArgs(factory, object, NULL);
@@ -1051,7 +1063,7 @@ lookup_adapter_hook(lookup *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &provided, &object, &name, &default_))
-    return NULL; 
+    return NULL;
 
   return _adapter_hook(self, provided, object, name, default_);
 }
@@ -1064,12 +1076,12 @@ lookup_queryAdapter(lookup *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &object, &provided, &name, &default_))
-    return NULL; 
+    return NULL;
 
   return _adapter_hook(self, provided, object, name, default_);
 }
 
-/*  
+/*
     def lookupAll(self, required, provided):
         cache = self._mcache.get(provided)
         if cache is None:
@@ -1124,7 +1136,7 @@ _lookupAll(lookup *self, PyObject *required, PyObject *provided)
       Py_DECREF(required);
     }
 
-  return result;  
+  return result;
 }
 static PyObject *
 lookup_lookupAll(lookup *self, PyObject *args, PyObject *kwds)
@@ -1134,12 +1146,12 @@ lookup_lookupAll(lookup *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist,
                                     &required, &provided))
-    return NULL; 
+    return NULL;
 
   return _lookupAll(self, required, provided);
 }
 
-/*  
+/*
     def subscriptions(self, required, provided):
         cache = self._scache.get(provided)
         if cache is None:
@@ -1195,7 +1207,7 @@ _subscriptions(lookup *self, PyObject *required, PyObject *provided)
       Py_DECREF(required);
     }
 
-  return result;  
+  return result;
 }
 static PyObject *
 lookup_subscriptions(lookup *self, PyObject *args, PyObject *kwds)
@@ -1205,25 +1217,24 @@ lookup_subscriptions(lookup *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist,
                                     &required, &provided))
-    return NULL; 
+    return NULL;
 
   return _subscriptions(self, required, provided);
 }
 
 static struct PyMethodDef lookup_methods[] = {
   {"changed",	    (PyCFunction)lookup_changed,       METH_O,        ""},
-  {"lookup",	    (PyCFunction)lookup_lookup,	       METH_KEYWORDS, ""},
-  {"lookup1",	    (PyCFunction)lookup_lookup1,       METH_KEYWORDS, ""},
-  {"queryAdapter",  (PyCFunction)lookup_queryAdapter,  METH_KEYWORDS, ""},
-  {"adapter_hook",  (PyCFunction)lookup_adapter_hook,  METH_KEYWORDS, ""},
-  {"lookupAll",	    (PyCFunction)lookup_lookupAll,     METH_KEYWORDS, ""},
-  {"subscriptions", (PyCFunction)lookup_subscriptions, METH_KEYWORDS, ""},
+  {"lookup",	    (PyCFunction)lookup_lookup,	       METH_KEYWORDS | METH_VARARGS, ""},
+  {"lookup1",	    (PyCFunction)lookup_lookup1,       METH_KEYWORDS | METH_VARARGS, ""},
+  {"queryAdapter",  (PyCFunction)lookup_queryAdapter,  METH_KEYWORDS | METH_VARARGS, ""},
+  {"adapter_hook",  (PyCFunction)lookup_adapter_hook,  METH_KEYWORDS | METH_VARARGS, ""},
+  {"lookupAll",	    (PyCFunction)lookup_lookupAll,     METH_KEYWORDS | METH_VARARGS, ""},
+  {"subscriptions", (PyCFunction)lookup_subscriptions, METH_KEYWORDS | METH_VARARGS, ""},
   {NULL,	    NULL}		/* sentinel */
 };
 
 static PyTypeObject LookupBase = {
-	PyObject_HEAD_INIT(NULL)
-	/* ob_size           */ 0,
+	PyVarObject_HEAD_INIT(NULL, 0)
 	/* tp_name           */ "_zope_interface_coptimizations."
                                 "LookupBase",
 	/* tp_basicsize      */ sizeof(lookup),
@@ -1232,7 +1243,7 @@ static PyTypeObject LookupBase = {
 	/* tp_print          */ (printfunc)0,
 	/* tp_getattr        */ (getattrfunc)0,
 	/* tp_setattr        */ (setattrfunc)0,
-	/* tp_compare        */ (cmpfunc)0,
+	/* tp_compare        */ 0,
 	/* tp_repr           */ (reprfunc)0,
 	/* tp_as_number      */ 0,
 	/* tp_as_sequence    */ 0,
@@ -1244,7 +1255,7 @@ static PyTypeObject LookupBase = {
         /* tp_setattro       */ (setattrofunc)0,
         /* tp_as_buffer      */ 0,
         /* tp_flags          */ Py_TPFLAGS_DEFAULT
-				| Py_TPFLAGS_BASETYPE 
+				| Py_TPFLAGS_BASETYPE
                           	| Py_TPFLAGS_HAVE_GC,
 	/* tp_doc            */ "",
         /* tp_traverse       */ (traverseproc)lookup_traverse,
@@ -1275,7 +1286,7 @@ verifying_traverse(verify *self, visitproc visit, void *arg)
     if (vret != 0)
       return vret;
   }
-  
+
   return 0;
 }
 
@@ -1293,10 +1304,10 @@ static void
 verifying_dealloc(verify *self)
 {
   verifying_clear(self);
-  self->ob_type->tp_free((PyObject*)self);
+  Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
-/*  
+/*
     def changed(self, originally_changed):
         super(VerifyingBasePy, self).changed(originally_changed)
         self._verify_ro = self._registry.ro[1:]
@@ -1307,13 +1318,13 @@ _generations_tuple(PyObject *ro)
 {
   int i, l;
   PyObject *generations;
-  
+
   l = PyTuple_GET_SIZE(ro);
   generations = PyTuple_New(l);
   for (i=0; i < l; i++)
     {
       PyObject *generation;
-      
+
       generation = PyObject_GetAttr(PyTuple_GET_ITEM(ro, i), str_generation);
       if (generation == NULL)
         {
@@ -1349,7 +1360,7 @@ verifying_changed(verify *self, PyObject *ignored)
   Py_DECREF(t);
   if (ro == NULL)
     return NULL;
-  
+
   self->_verify_generations = _generations_tuple(ro);
   if (self->_verify_generations == NULL)
     {
@@ -1363,7 +1374,7 @@ verifying_changed(verify *self, PyObject *ignored)
   return Py_None;
 }
 
-/*  
+/*
     def _verify(self):
         if ([r._generation for r in self._verify_ro]
             != self._verify_generations):
@@ -1383,16 +1394,17 @@ _verify(verify *self)
       if (generations == NULL)
         return -1;
 
-      changed = PyObject_Compare(self->_verify_generations, generations);
+      changed = PyObject_RichCompareBool(self->_verify_generations,
+					 generations, Py_NE);
       Py_DECREF(generations);
-      if (PyErr_Occurred())
+      if (changed == -1)
         return -1;
-      
+
       if (changed == 0)
         return 0;
     }
 
-  changed_result = PyObject_CallMethodObjArgs(OBJECT(self), strchanged, 
+  changed_result = PyObject_CallMethodObjArgs(OBJECT(self), strchanged,
                                               Py_None, NULL);
   if (changed_result == NULL)
     return -1;
@@ -1409,7 +1421,7 @@ verifying_lookup(verify *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &required, &provided, &name, &default_))
-    return NULL; 
+    return NULL;
 
   if (_verify(self) < 0)
     return NULL;
@@ -1425,7 +1437,7 @@ verifying_lookup1(verify *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &required, &provided, &name, &default_))
-    return NULL; 
+    return NULL;
 
   if (_verify(self) < 0)
     return NULL;
@@ -1441,7 +1453,7 @@ verifying_adapter_hook(verify *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &provided, &object, &name, &default_))
-    return NULL; 
+    return NULL;
 
   if (_verify(self) < 0)
     return NULL;
@@ -1457,7 +1469,7 @@ verifying_queryAdapter(verify *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO|OO", kwlist,
                                     &object, &provided, &name, &default_))
-    return NULL; 
+    return NULL;
 
   if (_verify(self) < 0)
     return NULL;
@@ -1473,7 +1485,7 @@ verifying_lookupAll(verify *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist,
                                     &required, &provided))
-    return NULL; 
+    return NULL;
 
   if (_verify(self) < 0)
     return NULL;
@@ -1489,7 +1501,7 @@ verifying_subscriptions(verify *self, PyObject *args, PyObject *kwds)
 
   if (! PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist,
                                     &required, &provided))
-    return NULL; 
+    return NULL;
 
   if (_verify(self) < 0)
     return NULL;
@@ -1499,18 +1511,17 @@ verifying_subscriptions(verify *self, PyObject *args, PyObject *kwds)
 
 static struct PyMethodDef verifying_methods[] = {
   {"changed",	   (PyCFunction)verifying_changed,	  METH_O,        ""},
-  {"lookup",	   (PyCFunction)verifying_lookup,	  METH_KEYWORDS, ""},
-  {"lookup1",	   (PyCFunction)verifying_lookup1,	  METH_KEYWORDS, ""},
-  {"queryAdapter",  (PyCFunction)verifying_queryAdapter,  METH_KEYWORDS, ""},
-  {"adapter_hook",  (PyCFunction)verifying_adapter_hook,  METH_KEYWORDS, ""},
-  {"lookupAll",	   (PyCFunction)verifying_lookupAll,	  METH_KEYWORDS, ""},
-  {"subscriptions", (PyCFunction)verifying_subscriptions, METH_KEYWORDS, ""},
+  {"lookup",	   (PyCFunction)verifying_lookup,	  METH_KEYWORDS | METH_VARARGS, ""},
+  {"lookup1",	   (PyCFunction)verifying_lookup1,	  METH_KEYWORDS | METH_VARARGS, ""},
+  {"queryAdapter",  (PyCFunction)verifying_queryAdapter,  METH_KEYWORDS | METH_VARARGS, ""},
+  {"adapter_hook",  (PyCFunction)verifying_adapter_hook,  METH_KEYWORDS | METH_VARARGS, ""},
+  {"lookupAll",	   (PyCFunction)verifying_lookupAll,	  METH_KEYWORDS | METH_VARARGS, ""},
+  {"subscriptions", (PyCFunction)verifying_subscriptions, METH_KEYWORDS | METH_VARARGS, ""},
   {NULL,	    NULL}		/* sentinel */
 };
 
 static PyTypeObject VerifyingBase = {
-	PyObject_HEAD_INIT(NULL)
-	/* ob_size           */ 0,
+	PyVarObject_HEAD_INIT(NULL, 0)
 	/* tp_name           */ "_zope_interface_coptimizations."
                                 "VerifyingBase",
 	/* tp_basicsize      */ sizeof(verify),
@@ -1519,7 +1530,7 @@ static PyTypeObject VerifyingBase = {
 	/* tp_print          */ (printfunc)0,
 	/* tp_getattr        */ (getattrfunc)0,
 	/* tp_setattr        */ (setattrfunc)0,
-	/* tp_compare        */ (cmpfunc)0,
+	/* tp_compare        */ 0,
 	/* tp_repr           */ (reprfunc)0,
 	/* tp_as_number      */ 0,
 	/* tp_as_sequence    */ 0,
@@ -1531,7 +1542,7 @@ static PyTypeObject VerifyingBase = {
         /* tp_setattro       */ (setattrofunc)0,
         /* tp_as_buffer      */ 0,
         /* tp_flags          */ Py_TPFLAGS_DEFAULT
-				| Py_TPFLAGS_BASETYPE 
+				| Py_TPFLAGS_BASETYPE
                           	| Py_TPFLAGS_HAVE_GC,
 	/* tp_doc            */ "",
         /* tp_traverse       */ (traverseproc)verifying_traverse,
@@ -1559,20 +1570,38 @@ static struct PyMethodDef m_methods[] = {
    "Get an object's interfaces (internal api)"},
   {"providedBy", (PyCFunction)providedBy, METH_O,
    "Get an object's interfaces"},
-  
+
   {NULL,	 (PyCFunction)NULL, 0, NULL}		/* sentinel */
 };
 
-#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
-#define PyMODINIT_FUNC void
+#if  PY_MAJOR_VERSION >= 3
+static char module_doc[] = "C optimizations for zope.interface\n\n";
+
+static struct PyModuleDef _zic_module = {
+	PyModuleDef_HEAD_INIT,
+	"_zope_interface_coptimizations",
+	module_doc,
+	-1,
+	m_methods,
+	NULL,
+	NULL,
+	NULL,
+	NULL
+};
 #endif
-PyMODINIT_FUNC
-init_zope_interface_coptimizations(void)
+
+static PyObject *
+init(void)
 {
   PyObject *m;
 
+#if  PY_MAJOR_VERSION < 3
 #define DEFINE_STRING(S) \
-  if(! (str ## S = PyString_FromString(# S))) return
+  if(! (str ## S = PyString_FromString(# S))) return NULL
+#else
+#define DEFINE_STRING(S) \
+  if(! (str ## S = PyUnicode_FromString(# S))) return NULL
+#endif
 
   DEFINE_STRING(__dict__);
   DEFINE_STRING(__implemented__);
@@ -1595,53 +1624,69 @@ init_zope_interface_coptimizations(void)
 #undef DEFINE_STRING
   adapter_hooks = PyList_New(0);
   if (adapter_hooks == NULL)
-    return;
-        
+    return NULL;
+
   /* Initialize types: */
   SpecType.tp_new = PyBaseObject_Type.tp_new;
   if (PyType_Ready(&SpecType) < 0)
-    return;
+    return NULL;
   OSDType.tp_new = PyBaseObject_Type.tp_new;
   if (PyType_Ready(&OSDType) < 0)
-    return;
+    return NULL;
   CPBType.tp_new = PyBaseObject_Type.tp_new;
   if (PyType_Ready(&CPBType) < 0)
-    return;
+    return NULL;
 
   InterfaceBase.tp_new = PyBaseObject_Type.tp_new;
   if (PyType_Ready(&InterfaceBase) < 0)
-    return;
+    return NULL;
 
   LookupBase.tp_new = PyBaseObject_Type.tp_new;
   if (PyType_Ready(&LookupBase) < 0)
-    return;
+    return NULL;
 
   VerifyingBase.tp_new = PyBaseObject_Type.tp_new;
   if (PyType_Ready(&VerifyingBase) < 0)
-    return;
+    return NULL;
 
-  
+  #if PY_MAJOR_VERSION < 3
   /* Create the module and add the functions */
   m = Py_InitModule3("_zope_interface_coptimizations", m_methods,
-                     "C optimizations for zope.interface\n\n"
-                     "$Id: _zope_interface_coptimizations.c 69150 2006-07-17 03:20:44Z fdrake $");  
+                     "C optimizations for zope.interface\n\n");
+  #else
+  m = PyModule_Create(&_zic_module);
+  #endif
   if (m == NULL)
-    return;
-  
+    return NULL;
+
   /* Add types: */
   if (PyModule_AddObject(m, "SpecificationBase", OBJECT(&SpecType)) < 0)
-    return;
-  if (PyModule_AddObject(m, "ObjectSpecificationDescriptor", 
+    return NULL;
+  if (PyModule_AddObject(m, "ObjectSpecificationDescriptor",
                          (PyObject *)&OSDType) < 0)
-    return;
+    return NULL;
   if (PyModule_AddObject(m, "ClassProvidesBase", OBJECT(&CPBType)) < 0)
-    return;
+    return NULL;
   if (PyModule_AddObject(m, "InterfaceBase", OBJECT(&InterfaceBase)) < 0)
-    return;
+    return NULL;
   if (PyModule_AddObject(m, "LookupBase", OBJECT(&LookupBase)) < 0)
-    return;
+    return NULL;
   if (PyModule_AddObject(m, "VerifyingBase", OBJECT(&VerifyingBase)) < 0)
-    return;
+    return NULL;
   if (PyModule_AddObject(m, "adapter_hooks", adapter_hooks) < 0)
-    return;
+    return NULL;
+  return m;
 }
+
+PyMODINIT_FUNC
+#if PY_MAJOR_VERSION < 3
+init_zope_interface_coptimizations(void)
+{
+  init();
+}
+#else
+PyInit__zope_interface_coptimizations(void)
+{
+  return init();
+}
+#endif
